@@ -1,5 +1,3 @@
-/* $Cambridge: exim/exim-src/src/spam.c,v 1.14 2007/05/14 18:56:25 magnus Exp $ */
-
 /*************************************************
 *     Exim - an Internet mail transport agent    *
 *************************************************/
@@ -20,15 +18,18 @@ uschar spam_report_buffer[32600];
 uschar prev_user_name[128] = "";
 int spam_ok = 0;
 int spam_rc = 0;
+uschar *prev_spamd_address_work = NULL;
 
-int spam(uschar **listptr) {
+int
+spam(uschar **listptr)
+{
   int sep = 0;
   uschar *list = *listptr;
   uschar *user_name;
   uschar user_name_buffer[128];
   unsigned long mbox_size;
   FILE *mbox_file;
-  int spamd_sock;
+  int spamd_sock = -1;
   uschar spamd_buffer[32600];
   int i, j, offset, result;
   uschar spamd_version[8];
@@ -46,6 +47,7 @@ int spam(uschar **listptr) {
   struct timeval select_tv;         /* and applied by PH */
   fd_set select_fd;
 #endif
+  uschar *spamd_address_work;
 
   /* stop compiler warning */
   result = 0;
@@ -70,6 +72,23 @@ int spam(uschar **listptr) {
     override = 1;
   };
 
+  /* expand spamd_address if needed */
+  if (*spamd_address == '$') {
+    spamd_address_work = expand_string(spamd_address);
+    if (spamd_address_work == NULL) {
+      log_write(0, LOG_MAIN|LOG_PANIC,
+        "spamassassin acl condition: spamd_address starts with $, but expansion failed: %s", expand_string_message);
+      return DEFER;
+    }
+  }
+  else
+    spamd_address_work = spamd_address;
+
+  /* check if previous spamd_address was expanded and has changed. dump cached results if so */
+  if ( spam_ok && ( prev_spamd_address_work != NULL) && (Ustrcmp(prev_spamd_address_work, spamd_address_work) != 0)) {
+    spam_ok = 0;
+  }
+
   /* if we scanned for this username last time, just return */
   if ( spam_ok && ( Ustrcmp(prev_user_name, user_name) == 0 ) ) {
     if (override)
@@ -79,7 +98,7 @@ int spam(uschar **listptr) {
   };
 
   /* make sure the eml mbox file is spooled up */
-  mbox_file = spool_mbox(&mbox_size);
+  mbox_file = spool_mbox(&mbox_size, NULL);
 
   if (mbox_file == NULL) {
     /* error while spooling */
@@ -89,14 +108,13 @@ int spam(uschar **listptr) {
   };
 
   start = time(NULL);
+
   /* socket does not start with '/' -> network socket */
-  if (*spamd_address != '/') {
-    time_t now = time(NULL);
+  if (*spamd_address_work != '/') {
     int num_servers = 0;
-    int current_server = 0;
-    int start_server = 0;
+    int current_server;
     uschar *address = NULL;
-    uschar *spamd_address_list_ptr = spamd_address;
+    uschar *spamd_address_list_ptr = spamd_address_work;
     uschar address_buffer[256];
     spamd_address_container * spamd_address_vector[32];
 
@@ -106,11 +124,13 @@ int spam(uschar **listptr) {
                                         address_buffer,
                                         sizeof(address_buffer))) != NULL) {
 
+      /* Potential memory leak as we never free the store. */
       spamd_address_container *this_spamd =
         (spamd_address_container *)store_get(sizeof(spamd_address_container));
 
       /* grok spamd address and port */
-      if( sscanf(CS address, "%s %u", this_spamd->tcp_addr, &(this_spamd->tcp_port)) != 2 ) {
+      if (sscanf(CS address, "%23s %u", this_spamd->tcp_addr, &(this_spamd->tcp_port)) != 2)
+        {
         log_write(0, LOG_MAIN,
           "spam acl condition: warning - invalid spamd address: '%s'", address);
         continue;
@@ -130,9 +150,11 @@ int spam(uschar **listptr) {
       return DEFER;
     };
 
-    current_server = start_server = (int)now % num_servers;
+    while ( num_servers > 0 ) {
+      int i;
 
-    while (1) {
+      /* Randomly pick a server to try */
+      current_server = random_number( num_servers );
 
       debug_printf("trying server %s, port %u\n",
                    spamd_address_vector[current_server]->tcp_addr,
@@ -160,16 +182,20 @@ int spam(uschar **listptr) {
          spamd_address_vector[current_server]->tcp_addr,
          spamd_address_vector[current_server]->tcp_port,
          strerror(errno));
-      current_server++;
-      if (current_server >= num_servers)
-        current_server = 0;
-      if (current_server == start_server) {
-        log_write(0, LOG_MAIN|LOG_PANIC, "spam acl condition: all spamd servers failed");
-        (void)fclose(mbox_file);
-        (void)close(spamd_sock);
-        return DEFER;
-      };
-    };
+
+      (void)close(spamd_sock);
+
+      /* Remove the server from the list. XXX We should free the memory */
+      num_servers--;
+      for( i = current_server; i < num_servers; i++ )
+        spamd_address_vector[i] = spamd_address_vector[i+1];
+    }
+
+    if ( num_servers == 0 ) {
+      log_write(0, LOG_MAIN|LOG_PANIC, "spam acl condition: all spamd servers failed");
+      (void)fclose(mbox_file);
+      return DEFER;
+    }
 
   }
   else {
@@ -184,12 +210,12 @@ int spam(uschar **listptr) {
     }
 
     server.sun_family = AF_UNIX;
-    Ustrcpy(server.sun_path, spamd_address);
+    Ustrcpy(server.sun_path, spamd_address_work);
 
     if (connect(spamd_sock, (struct sockaddr *) &server, sizeof(struct sockaddr_un)) < 0) {
       log_write(0, LOG_MAIN|LOG_PANIC,
                 "malware acl condition: spamd: unable to connect to UNIX socket %s (%s)",
-                spamd_address, strerror(errno) );
+                spamd_address_work, strerror(errno) );
       (void)fclose(mbox_file);
       (void)close(spamd_sock);
       return DEFER;
@@ -197,6 +223,14 @@ int spam(uschar **listptr) {
 
   }
 
+  if (spamd_sock == -1) {
+    log_write(0, LOG_MAIN|LOG_PANIC,
+        "programming fault, spamd_sock unexpectedly unset");
+    (void)fclose(mbox_file);
+    (void)close(spamd_sock);
+    return DEFER;
+  }
+
   /* now we are connected to spamd on spamd_sock */
   (void)string_format(spamd_buffer,
            sizeof(spamd_buffer),
@@ -255,7 +289,7 @@ again:
             "spam acl condition: %s on spamd socket", strerror(errno));
         else {
           if (time(NULL) - start < SPAMD_TIMEOUT)
-          goto again;
+            goto again;
           log_write(0, LOG_MAIN|LOG_PANIC,
             "spam acl condition: timed out writing spamd socket");
         }
@@ -341,13 +375,10 @@ again:
     *q = *p;
     q++;
     if (*p == '\n') {
-      *q = '\t';
+      /* add an extra space after the newline to ensure
+      that it is treated as a header continuation line */
+      *q = ' ';
       q++;
-      /* eat whitespace */
-      while( (*p <= ' ') && (*p != '\0') ) {
-        p++;
-      };
-      p--;
     };
     p++;
   };
@@ -395,6 +426,10 @@ again:
     spam_rc = FAIL;
   };
 
+  /* remember expanded spamd_address if needed */
+  if (spamd_address_work != spamd_address) {
+    prev_spamd_address_work = string_copy(spamd_address_work);
+  }
   /* remember user name and "been here" for it */
   Ustrcpy(prev_user_name, user_name);
   spam_ok = 1;
